#!/usr/bin/env python3
'''Provide generic infrastructure for models with parameters

Relations contain organized sets of parameters.

'''
from gwalk.model.parameter import Parameter
class Relation(object):
    '''Base class for containing parameters and relationships among them

    Ideally, these will be additive,
        which means we must use a dictionary rather than a list

    '''
    def __init__(
                 self,
                 parameters=None,
                 random_state=None,
                ):
        '''Initialize model with some parameters
        
        Parameters
        ----------
        parameters: list, optional
            Input A list, tuple, or dict of parameters
                These are optional, and can be added afterwards
        random_state: np.RandomState, optional
            A random state for random number generation
        '''
        ## Imports
        # Public
        import numpy as np
        # Initialize random state
        if random_state is None:
            self.rs = np.random.RandomState()
        else:
            self.rs = random_state

        # Initialize self._parameters
        self._parameters = {}

        # determine if there are any starting parameters
        if not (parameters is None):
            if isinstance(parameters, dict):
                for key in parameters:
                    self.add_parameter(parameters[key])
            if isinstance(parameters, list) or isinstance(parameters, tuple):
                for item in parameters:
                    self.add_parameter(item)
            elif isinstance(parameters, Parameter):
                self.add_parameter(parameters)
            else:
                raise TypeError("Unknown Parameter type: %s"%(str(type(parameters))))


    def add_parameter(self, p):
        '''Append a parameter to our list of parameters

        Parameters
        ----------
        p: Parameter
            Input new parameter object to append to relation
        '''
        if p.name in self._parameters:
            raise RuntimeError(
                    "Failed to add %s! Parameters must be unique!"%(p.name)
                              )
        self._parameters[p.name] = p
        self._parameters[p.name].rs = self.rs
        # Update parameter map
        self.map_parameters()

    def assign_guess(self, X):
        '''Assign guesses to each parameter

        Parameters
        ----------
        X: array like, shape = nd
            An array of guesses for our parameters,
        '''
        # Check that sample is valid
        X = self.check_sample(X)

        # Initialize constrained Samples
        ns, nd = X.shape

        # Assert things about the shape
        assert ns == 1
        assert nd == len(self._parameters)

        # Check if our guess makes sense
        assert self.satisfies_constraints(X)

        # Assign the guess
        for i, key in enumerate(self._parameters):
            # Do it
            self._parameters[key].assign_guess(X[0,i])

    def read_guess(self):
        '''Read the current guess'''
        ## Imports ##
        # Public
        import numpy as np
        guess = []
        for key in self._parameters:
            guess.append(self._parameters[key].guess)
        return np.asarray(guess)

    def read_limits(self):
        '''Read limits of parameters'''
        ## Imports ##
        # Public
        import numpy as np
        limits = []
        for key in self._parameters:
            limits.append(self._parameters[key].limits)
        return np.asarray(limits)

    def update_limits(self, limits):
        '''Update the limits in an intelligent way

        Parameters
        ----------
        limits: array like, shape = (nd, 2)
            Input a new list of limits [min,max] for each parameter
        '''
        ## Imports
        # Public
        import numpy as np
        # Check dimensionality of limits
        if not np.asarray(limits).shape == (len(self._parameters),2):
            raise RuntimeError("limits shape is %s, but should be %s"%(
                str(np.asarray(limits.shape)), str(len(self._parameters),2)))
        for i, key in enumerate(self._parameters):
            self._parameters[key].update_limits(limits[i])

    def map_parameters(self):
        '''Generate a map of parameters to linear indicies'''
        # Next, get a parameter dictionary
        param_ids = {}
        for i, key in enumerate(self._parameters):
            param_ids[key] = i

        # Hold onto that dictionary
        self.p_map = param_ids

        return self.p_map

        
    def satisfies_constraints(self, X):
        '''Check if samples satisfy constraints

        By default there are no constraints on a relationship
            apart from parameter constraints

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input an array of samples which may describe 
                the parameters of this relationship.
        '''
        ## Imports 
        # Public
        import numpy as np
        
        # Check that samples are valid
        X = self.check_sample(X)

        # Initialize constrained Samples
        ns, nd = X.shape
        constrained_indx = np.ones(ns).astype(bool)

        # Constrain by each parameter
        for i, key in enumerate(self._parameters):
            # Identify the parameter index
            p_indx = self._parameters[key].satisfies_constraints(X[:,i])
            # Bitwise and with relationship constraints
            constrained_indx = constrained_indx & p_indx

        return constrained_indx

    # Draw random samples
    def sample_uniform_unconstrained(self, samples):
        '''Sample all parameters uniformly

        Parameters
        ----------
        samples: int
            Input the number of uniform samples we would like to draw
        '''
        ## Imports
        # Public
        import numpy as np
        # Initialize sample shape
        sample_shape = (samples, len(self._parameters.keys()))
        # Initialize samples
        Xs = np.zeros(sample_shape)

        # Fill in values
        for i, key in enumerate(self._parameters):
            Xs[:,i] = self._parameters[key].sample_uniform(samples)

        return Xs

    # Draw random samples in a constrained way
    def sample_uniform(self, samples):
        '''Sample parameters, being careful with constraints

        Parameters
        ----------
        samples: int
            Input the number of uniform samples we would like to draw
        '''
        ## Imports
        # Public
        import numpy as np

        # Initialize the number of valid samples we have
        n_keep = 0

        # Initialize sample shape
        sample_shape = (samples, len(self._parameters.keys()))
        # Initialize samples to keep
        Xk = np.zeros(sample_shape)

        # Continue doing this until we have the right number of samples
        while n_keep < samples:
            # How many samples do we still need
            n_need = samples - n_keep
            # Draw that many samples
            Xs = self.sample_uniform_unconstrained(n_need)
            # Check if these samples satisfy our constraints
            k_indx = self.satisfies_constraints(Xs)
            # Identify the number of new samples to keep from this iteration
            n_it = np.sum(k_indx)
            # keep iterating if there are no valid samples
            if n_it != 0:
                # Keep valid samples
                Xs = Xs[k_indx]
                # Add them to Xk
                Xk[n_keep:n_keep+n_it,:] = Xs
                # Updated n_keep
                n_keep += n_it

        # Return samples
        return Xk

    def check_sample(
                       self,
                       X,
                       w = None,
                       *args,
                       **kwargs
                     ):
        '''Check dimensionality of samples
        
        This function is setup for inheirtance.

        Other relation objects may not have methods this simple

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample we would like to check as a valid sample
        w: array like, shape = (npts)
            Input weights for valid sample
        '''
        ## Imports ##
        # Public
        import numpy as np
        import warnings

        # Make sure data is the right shape
        if not (len(X.shape) == 2):
            # Check if the data is only one dimensional
            if len(X.shape) == 1:
                # We can make this two dimensional
                X = np.atleast_2d(X)
            else:
                # The data must be two dimensional
                raise RuntimeError(
                    "X.shape = %s, model has %d parameters!"%(str(X.shape), nd)
                    )

        # Check the dimensionality of our parameters
        nd = len(self._parameters)

        # Make sure data has the right dimensions
        if not X.shape[1] == nd:
            # Check if the data just needs to be transposed
            if X.shape[0] == nd:
                warnings.warn(
                    "X.shape = %s, model has %d parameters!"%(str(X.shape), nd)
                    )
                # Transpose data
                X = X.T
            else:
                raise RuntimeError(
                        "X.shape = %s, model has %d parameters!"%(str(X.shape), nd)
                        )

        # Check weights
        if not (w is None):
            # Check that weights have the right shape
            if not len(w.shape) == 1:
                raise RuntimeError(
                    "Weights has shape %s, but should be one dimensional"%(
                     str(w.shape)),
                    )

            # Check number of weights
            if not (x.shape[0] == w.size):
                # We have a different number of weights and samples
                raise RuntimeError(
                    "X.shape = %s, w.shape = %s"%(str(X.shape), str(w.shape))
                   )

        return X


    def likelihood(
                   self,
                   X,
                   return_product=False,
                   log_scale=False,
                   *args,
                   **kwargs
                  ):
        '''Placeholder function for likelihood
        
        In general, relationship objects will have more complicated 
            likelihood functions

        Parameters
        ----------
        X: array like (npts, nd)
            Input sample we would like to evaluate
        return_product: bool, optional
            Input return product of values instead of array?
        log_scale: bool, optional
            Input return lnL instead of L
        '''
        ## Imports ##
        # Public
        import numpy as np
        import warnings

        # Check that the samples make sense
        X = self.check_sample(X)

        # Check the number of samples being evaluated
        ns, nd = X.shape

        # This is only a placeholder likelihood function
        L = np.ones((ns,))

        # Check if we want the log scale
        if log_scale and return_prod:
            return np.sum(np.log(L))
        elif log_scale:
            return np.log(L)
        elif return_prod:
            return np.prod(L)
        else:
            return L

    def hypercube(
                  self,
                  res=None,
                  size=None,
                  limits=None,
                  constrained=False,
                 ):
        '''Generate a hypercube in all of our parameters

        Parameters
        ----------
        res: array like, shape = (nd), optional
            Input Desired hypercube resolution in each dimension
        size: int, optional
            Input maximum number of hypercube points
        limits: array like, shape = (nd, 2), optional
            Alternative limits for hypercube
        '''
        ## Imports ##
        # Public
        import numpy as np
        from gp_api.utils import sample_hypercube
        import warnings

        ## Initialize resolution ##
        # Check dimensionality
        ndim = len(self._parameters)

        # Check cases
        if (res is None):
            if (size is None):
                raise RuntimeError("Usage: hypercube must specify res or size")
            else:
                # Generate maximum resolution for given size
                res = int(size**(1/ndim))
                # This is not a very dense hypercube if res < 5
                if res < 5:
                    warnings.warn(
                        "Maximum res for %d-D hypercube wiht size %d is %d"%(
                            (ndim, size, res))
                                 )
        elif type(res) == int:
            if (size is None):
                # Make res an array
                res = (res*np.ones(ndim)).astype(int)
            else:
                if res > int(size**(1/ndim)):
                    raise RuntimeError(
                        "resolution and size both specified and conflict!")
                else:
                    # Make res an array
                    res = (res*np.ones(ndim)).astype(int)
        elif not(type(res) in [list, tuple, np.ndarray]):
            raise RuntimeError("Unknown resolution type %s"%(str(type(res))))
        elif not (len(res) == ndimn):
            raise RuntimeError("Resolution is not the correct length")
        else:
            # Make res an array
            res = np.asarray(res).astype(int)
            

        ## Initialize limits ##
        if limits is None:
            limits = self.read_limits()
        else:
            # Make limits an array
            limits = np.asarray(limits)
            # Check that limits make sense
            if not limits.shape == (ndim, 2):
                raise RuntimeError("Strange limits fed to hypercube: %s"%(
                        str(limits)))

        ## Generate hypercube ##
        cube_space = sample_hypercube(limits, res)

        # Check where constraints are satisfied
        if constrained:
            k_indx = self.satisfies_constraints(cube_space)
            cube_space = cube_space[k_indx]

        return cube_space
                    
    def __add__(self, other):
        '''Add two relationships by combining parameters and constraints

        Parameters
        ----------
        other: Relationship
            Input: other relationship you wish to add
        '''
        new = _AddRelation(self, other)

        return new

class _AddRelation(Relation):
    '''This is a relationship that results from adding other relationships

    It should not be initialized by a user
    '''
    def __init__(self, summand1, summand2):
        '''Initialize the added parameters

        Parameters
        ----------
        summand1: Relation
            Input first summand for add relation
        summand2: Relation
            Input second summand for add relation
        '''
        # Initialize random state
        self.rs = summand1.rs
        # Hold summands
        self.summands=[]
        # Update summands with summand 1
        if isinstance(summand1, _AddRelation):
            for item in summand1.summands:
                self.summands.append(item)
        else:
            self.summands.append(summand1)

        # Update summands with summand 2
        if isinstance(summand2, _AddRelation):
            for item in summand2.summands:
                self.summands.append(item)
        else:
            self.summands.append(summand2)

        # Define paremeters
        self._parameters = {}
        for summand in self.summands:
            for item in summand._parameters:
                self.add_parameter(summand._parameters[item])

            

    def satisfies_constraints(self, X):
        '''Check if add relation satisfies constraints

        This part is a little sneaky

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample to evaluate constraints on
        '''
        ## Imports ##
        # Public
        import numpy as np

        # Identify information
        ns, nd = X.shape
        # Initialize Constrained Index 
        constrained_indx = np.ones(ns).astype(bool)
        
        ## Loop the summands ##
        for summand in self.summands:
            # Identify the number of parameters
            n_param = len(summand.p_map.keys())
            # Initialize an array for parameter values
            Xp = np.zeros((ns, n_param))
            # Fill that array
            for key in summand.p_map.keys():
                Xp[:,summand.p_map[key]] = X[:,self.p_map[key]]
            # Update summand parameters
            for key in summand.p_map.keys():
                summand._parameters[key].update_limits(self._parameters[key].limits)
                summand._parameters[key].assign_guess(self._parameters[key].guess)

            # Calculate summand constraints
            constrained_indx &= summand.satisfies_constraints(Xp)

        # Return constrained indx
        return constrained_indx



#!/usr/bin/env python3
'''\
A layer of wrappers for an hdf5 database

database.py

Vera Del Favero

This database class will handle all access to the hdf5 databses
    we'll be using in this module
'''
######## Imports ########

import numpy as np
import time
import h5py
from os.path import join, isfile, isdir

######## Database Class ########

class Database(object):
    '''\
    This is a wrapper for hdf5 databases
        inteded to simplify calling and reduce clutter

    Inputs:
        fname (str): file name and location
        group (str): point to a particular group in the database

    Methods: 
        #### General ####

        change_group: change the group the object points to
        visit: visit each item in the databse
        merge: merge a similar database into this one
        shard: create a new databse with some data from the current database
        list_items: list databases and groups in a particular group within
            the hdf5 database
        scan: scan a particular group in the database and print it
            without changing the group we point to
        kind: determine if a particular item is a database or a group
        exists: determine if a particular item exists

        #### dset methods Methods ####

        dset_init: initialize new dataset in the hdf5 database
        dset_size: return the size of the dataset
        dset_compression: return the compression of the dataset
        dset_shape: return the shape of the dataset
        dset_dtype: return the dtype of the dataset
        dset_fields: return the fields for a dataset stored with fields
        dset_value: return the value of a dataset, 
            for a given field and set of indicies
        dset_sum: return the sum of a given dataset for some field 
            on some indicies
        dset_min: return the mix of a given dataset for some field 
            on some indicies
        dset_max: return the max of a given dataset for some field 
            on some indicies
        dset_set: set the values of a dataset, with some intelligence
        dset_rm: remove a given dataset (this does not free the disk space)
        dset_copy: copy a given dataset
        dset_recompress: copy a dataset to a temporary location and
            copy it back, changing the compression.
            WARNING: This is extraordinarily inneficient, and should only
            be used for testing

        #### attr methods ####

        attr_exists: check if attribute exists
        attr_value: return attribute value
        attr_set: set a given attribute
        attr_list: list attributes for a given dataset or group
        attr_dict: return the dictionary of attributes for a given dataset
            or group
        attr_set_dict: set each attribute from a given dictionary

    '''

    def __init__(
                 self, 
                 fname,
                 group = '/',
                 retries = 10,
                 sleep = 5,
                ):
        '''\
        Point to and ensure the existence of the database
    
        Inputs:
            fname (str): file name and location
            group (str): point to a particular group in the database

        '''
        # Assign attributes
        self.fname = fname
        if group is None:
            group = '/'
        self.group = group
        self.retries = retries
        self.sleep = sleep

        # Assert the existence of the database
        try:
            # First write a new database if one doesn't already exist
            with h5py.File(self.fname, 'w-') as F:
                if not (group is '/'):
                    F.create_group(group)
            return
        except:
            pass

        # Otherwise try to load the existing database
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r+') as F:
                    load = True
                    if (not (group is '/')) and (not (group in F)):
                        F.create_group(group)
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)


    def change_group(
                     self,
                     group = '/',
                    ):
        '''\
        Recreate the database with new group

        Inputs:
            group: New group
        '''

        # Check if group startswith /
        if group.startswith("/"):
            # group is to be interpreted global
            pass
        else:
            # Group is to be interpreted local
            group = join(self.group, group)

        # Create a new database object and assign it to this group
        self = Database(self.fname, group)

    def visit(self, fn = print):
        '''\
        Visit the database
        
        Inputs:
            fn: function to visit every item in database with
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    F.visit(fn)
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

    def merge(self, fname_head):
        '''\
        Merge another similar database (the head) into this one (the origin)

            Inputs:
                fname_head: the file name and location for the head node
        '''

        # Initialize two database objects
        Origin = h5py.File(self.fname, 'r+')
        Head = h5py.File(fname_head, 'r')

        # Make use of visit functionality to recursively copy database
        def visit_function(parcel):
            item = Head[parcel]
            # If the group or dataset is not in origin, copy it
            if not (item.name in Origin):
                print("found it!")
                Head.copy(item.name, Origin[item.parent.name])
            # If the group or dataset has attributes, copy them
            attrs = dict(item.attrs)
            # For each attr in the dict
            for key in list(attrs.keys()):
                # Copy it
                Origin[item.parent.name].attrs[key] = attrs[key]

        try:
            Head.visit(visit_function)

        finally:
            Origin.close()
            Head.close()

    def shard(self, fname_head, names = None, compression = "default"):
        '''\
        Create a smaller database with less data

        option to compress data differently
        
        Inputs:
            fname_head: file name and location for desired shard database
            names: list of items for copying
            compression: file compression option for database
        '''
        # Make sure we don't delete everything
        if self.fname == fname_head:
            raise Exception("shard name cannot be origin name")

        # Create objects for current database (origin)
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as Origin:
                    # and new database (head)
                    with h5py.File(fname_head, 'w') as Head:
                        load = True
                        # If names is not given, assume all data is to be copied
                        if names is None:
                            names = self.list_items(kind = "dset")
                        print("creating shard with members: ", names)


                        # Make sure the group exists in the new dataset
                        if not self.group == "/":
                            Head.create_group(self.group)

                        # Copy group attributes
                        group_attrs = self.attr_list('.')
                        for item in group_attrs:
                            Head[self.group].attrs[item] = Origin[self.group].attrs[item]

                        # Loop through each dataset
                        for item in names:
                            # Find the data
                            dset = Origin[self.group][item]
                            # Check compression
                            if compression =="default" or dset.size <= 1:
                                # Default copy behavior includes attrs
                                Origin.copy(dset.name, Head[self.group])
                            else:
                                # Copy data
                                Head[self.group].create_dataset(
                                                                item,
                                                                data=dset[...],
                                                                dtype=dset.dtype,
                                                                compression=compression
                                                               )
                                # copy attrs
                                for jtem in dset.attrs:
                                    Head[self.group][item].attrs[jtem] = dset.attrs[jtem]
    
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

            print("Shard successful!")

    def list_items(self, path="./", kind = None):
        '''\
        Return list of groups in current group
        
        Inputs:
            path: Path to group inside database.
                can be relative or absolute
            kind: list only datasets or only groups
        '''
        # Open File
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    # Find your place in the file
                    group_obj = F[self.group][path]
                    # Initialize list of items
                    items = []
                    # Go through current 'directory'
                    for item in group_obj:
                        # Keep track of what you want to append
                        append = False
                    
                        if kind is None:
                            # If kind is None, always append item
                            append = True

                        elif kind == "group":
                            # If kind is group, append it only if it is a group
                            if type(group_obj[item]) == h5py._hl.group.Group:
                                append = True

                        elif kind == "dset":
                            # If kind is dset, append only if it is a dset
                            if type(group_obj[item]) == h5py._hl.dataset.Dataset:
                                append = True

                        if append:
                            items.append(item)

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return items

    def scan(self, path="./", kind=None):
        '''\
        Return basic information about the objects in the current group

        This is really useful when exploring an unknown database
        Inputs:
            path: Path to group inside database.
                can be relative or absolute
            kind: list only datasets or only groups
        '''
        # Create temporary database object pointing at path
        db_scan = Database(self.fname, self.group)
        db_scan.change_group(path)
        # Read items in scanned pointer database
        items = db_scan.list_items(kind=kind)
        # Check each item
        for item in items:
            # Item information
            item_dict = {}
            # Name each item
            item_dict["name"] = item
            # Figure out if it is a dset or grou
            item_dict["kind"] = db_scan.kind(item)
            # List attributes
            item_dict["attrs"] = db_scan.attr_list(item)
            # Dset only things
            if item_dict["kind"] == "dset":
                # Return the shape of the dset
                item_dict["shape"] = db_scan.dset_shape(item)
                # Return the dtype of the dset
                item_dict["dtype"] = db_scan.dset_dtype(item)
                # Return the compression of the dset
                item_dict["compression"] = db_scan.dset_compression(item)
                # Return the fields of the dset
                item_dict["fields"] = db_scan.dset_fields(item)

            # Print item_dict
            print(item_dict)


    def create_group(self, item, clean = True):
        '''\
        Create a new group
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r+') as F:
                    load = True
                    group_obj = F[self.group]
                    if not item in group_obj:
                        group_obj.create_group(item)
                    else:
                        if clean:
                            raise RuntimeError("item %s already exists!"%(item))
                        if type(group_obj[item]) != h5py._hl.group.Group:
                            raise RuntimeError("item %s already exists, and is not a group!"%(item))
    
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return Database(self.fname, join(self.group, item))

    def kind(self, item):
        '''\
        Return the kind of item, be it a group or dset
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    if not (item in group_obj):
                        raise RuntimeError("kind: No such object %s"%(join(self.group, item)))
                    elif type(group_obj[item]) == h5py._hl.group.Group:
                        return "group"
                    elif type(group_obj[item]) == h5py._hl.dataset.Dataset:
                        return "dset"

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

    def exists(self, item, kind=None):
        '''\
        Check if item exists
        '''

        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    if not (item in group_obj):
                        return False
                    elif kind is None:
                        return True
                    elif kind == self.kind(item):
                        return True
                    else: return False

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)


    def dset_init(self, item, shape, dtype, compression=None ):
        '''\
        Initialize a new dataset

        Inputs:
            item: name of dataset
            shape: shape of dataset
            dtype: dtype of dataset
            compression: desired compression of data
        '''
        # Make the syntax a little more forgiving
        if not type(shape) is tuple:
            shape = tuple([shape])

        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r+') as F:
                    load = True
                    group_obj = F[self.group]
                    group_obj.create_dataset(item, shape, compression=compression, dtype=dtype)

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)


    def dset_size(self, item):
        '''\
        Return the size of a dataset

        Inputs:
            item: the dataset we would like to know the size of
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].size

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def dset_compression(self, item):
        '''\
        Return the compression of a dataset

        Inputs:
            item: the dataset we would like to know the size of
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].compression

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def dset_shape(self, item):
        '''\
        Return the sum of a dataset

        Inputs:
            item: the dataset we would like to know the size of
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].shape

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def dset_dtype(self, item):
        '''\
        Return the dtype of a dataset

        Inputs:
            item: the dataset we would like to know the size of
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].dtype

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def dset_fields(self, item):
        '''\
        Return the fields of a dataset with an informative dtype

        Inputs:
            item: the dataset we would like to know the size of
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].dtype.fields

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def dset_value(self, item, field = None, samples = None):
        '''\
        Return the values of a dataset

        Inputs:
            item: the dataset we would like to know the size of
            field: the field of interest for a complex data type
            samples: an array of indexes pointing to which part of the set
                we would like to analyze

        '''

        # Check usage
        if (field is None) and not (self.dset_fields(item) is None):
            raise ValueError("User did not specify a field.")

        # Otherwise open the file, and get the values
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    # Define the group
                    group_obj = F[self.group]

                    # find the appropriate samples
                    if samples is None:
                        value = group_obj[item][...]
                    else:
                        value = group_obj[item][samples]

                    # Find the appropriate field
                    if not (field is None):
                        value = value[field]

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value


    def dset_sum(self, item, field = None, samples = None):
        '''\
        Return the sum of a dataset

        Inputs:
            item: the dataset we would like to know the size of
            field: the field of interest for a complex data type
            samples: an array of indexes pointing to which part of the set
                we would like to analyze
        '''
        
        # Check usage
        if (field is None) and not (self.dset_fields(item) is None):
            raise ValueError("User did not specify a field.")

        if field is None:
            attr_name = "sum"
        else:
            attr_name = "%s_sum"%(field)

        # Check if the sum was already computed
        if self.attr_exists(item, attr_name):
            value = self.attr_value(item, attr_name)
        else:
            value = None

        # Case 1: precomputed sum
        if (not (value is None)) & (samples is None):
            return value

        # Case 2: calculate sum
        else:
            # Find the data
            data = self.dset_value(item, field = field, samples = samples)

            # Calcualte the sum
            value = np.sum(data)

        # Only save the sum if using the whole dataset
        if (samples is None):
            # Save the value for next time
            self.attr_set(item, attr_name, value)

        return value

    def dset_min(self, item, field = None, samples = None):
        '''\
        Return the minimum value of a dataset

        Inputs:
            item: the dataset we would like to know the size of
            field: the field of interest for a complex data type
            samples: an array of indexes pointing to which part of the set
                we would like to analyze
        '''

        # Check usage
        if (field is None) and not (self.dset_fields(item) is None):
            raise ValueError("User did not specify a field.")

        # Find the right attribute
        if field is None:
            attr_name = "min"
        else:
            attr_name = "%s_min"%(field)

        # Case 1: Check if the min was already computed
        if (samples is None) & self.attr_exists(item, attr_name):
            value = self.attr_value(item, attr_name)
 
        # Case 2: calculate min
        else:
            # Find the data
            data = self.dset_value(item, field = field, samples = samples)

            # Calcualte the minimum
            value = np.min(data)

        # Save the minimum if using the whole dataset
        if (samples is None):
            # Save the value for next time
            self.attr_set(item, attr_name, value)
        # Also save the minimum if it doesn't exist
        elif not (self.attr_exists(item, attr_name)):
            self.attr_set(item, attr_name, value)
        # Also save the minimum if it is less than the saved minimum
        elif self.attr_value(item, attr_name) > value:
            self.attr_set(item, attr_name, value)
        # Also return the saved value if it is less than the calculated value
        # Not sure if I really want this behavior.
        # It could be misleading
        else:
            value = self.attr_value(item, attr_name)

        return value

    def dset_max(self, item, field = None, samples = None):
        '''\
        Return the maximum value of a dataset
        '''

        # Check usage
        if (field is None) and not (self.dset_fields(item) is None):
            raise ValueError("User did not specify a field.")

        # Find the right attribute
        if field is None:
            attr_name = "max"
        else:
            attr_name = "%s_max"%(field)

        # Case 1: Check if the max was already computed
        if (samples is None) & self.attr_exists(item, attr_name):
            value = self.attr_value(item, attr_name)
 
        # Case 2: calculate max
        else:
            # Find the data
            data = self.dset_value(item, field = field, samples = samples)

            # Calcualte the minimum
            value = np.max(data)

        # Save the maximum if using the whole dataset
        if (samples is None):
            # Save the value for next time
            self.attr_set(item, attr_name, value)
        # Also save the maximum if it doesn't exist
        elif not (self.attr_exists(item, attr_name)):
            self.attr_set(item, attr_name, value)
        # Also save the maximum if it is greater than the saved value
        elif self.attr_value(item, attr_name) < value:
            self.attr_set(item, attr_name, value)
        # Return the saved value if it is greater than the calculated value
        # Not sure if I really want this behavior.
        # It could be misleading
        else:
            value = self.attr_value(item, attr_name)

        return value

    def dset_set(self, item, data, samples=None, fields=None, compression=None):
        '''\
        Create a new dataset or update the existing one

        Inputs:
            item: the dataset we would like to know the size of
            data: the data belonging to the dataset

            samples: an array of indexes pointing to which part of the set
                we would like to analyze
            fields: fields for complex datatype
            compression: Desired data compression
        '''

        if not (fields is None):
            raise NotImplementedError(
                "Reading from fields is fine, but we're not writing them yet!"
               )

        # Guarentee existence
        if not self.exists(item):
            if samples is None:
                self.dset_init(item, data.shape, data.dtype, compression=compression)
            else:
                raise Exception(
                    "Cannot initialize database for incomplete dataset!"
                   )
        else:
            if not (compression is None):
                raise KeyError("Can only set compression for new datasets")

        # Provide values
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r+') as F:
                    load = True
                    group_obj = F[self.group]
                    if (samples is None):
                        group_obj[item][...] = data
                    else:
                        group_obj[item][samples] = data
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)


    def dset_rm(self, item):
        '''\
        Remove a dataset
        '''
        if self.exists(item):
            n_try = 0
            load = False
            while (not load) and (n_try < self.retries):
                try:
                    with h5py.File(self.fname, 'r+') as F:
                        load = True
                        group_obj = F[self.group]
                        del group_obj[item]
                except OSError:
                    n_try += 1
                    #If the resource is busy, try again!
                    time.sleep(np.random.random()*self.sleep)

        else:
            raise Warning("Couldn't find %s to delete it!"%item)

    def dset_copy(self, item, copy, compression="default", dtype="default"):
        '''\
        Copy a dataset

        Inputs:
            item: the name of the item we would like to copy
            copy: the destination of the copy
            compression: the desired compression level of the copy
            dtype: the desired datatype for the new data
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, "r+") as F:
                    load = True
                    # Locate database
                    dset = F[self.group][item]
                    # Simple copy or recompress
                    if compression =="default" and dtype == "default":
                        # Default copy behavior includes attrs
                        F.copy(dset.name, F[self.group], name=copy)
                    else:
                        # Check which dtype to use
                        if dtype == "default":
                            dtype = dset.dtype
                        # Copy data
                        F[self.group].create_dataset(
                                                 copy,
                                                 data=dset[...],
                                                 dtype=dtype,
                                                 compression=compression,
                                                    )
                        for jtem in dset.attrs:
                            F[self.group][copy].attrs[jtem] = dset.attrs[jtem]
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

    
    def dset_recompress(self, dset, compression="gzip", dtype="default", alias="TEMP"):
        '''\
        Compress and copy data, and recopy it back to the same name

        This is actually much less useful than it sounds,
            and should only be used for testing
        '''
        self.dset_copy(dset, alias, compression="default", dtype="default")
        self.dset_rm(dset)
        self.dset_copy(alias, dset, compression=compression, dtype=dtype)
        self.dset_rm(alias)


    def attr_exists(self, item, attr):
        '''\
        Check if given attribute exists

        Inputs:
            item: the dataset or group the attribute would belong to
            attr: the name of the attribute
        '''

        n_try = 0
        load = False
        if not self.exists(item):
            raise Exception ("There is no such group %s"%item)
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    truth = attr in group_obj[item].attrs
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return truth

    def attr_value(self, item, attr):
        '''\
        Get the value of an attribute

        Inputs:
            item: the dataset or group the attribute would belong to
            attr: the name of the attribute
        '''

        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r') as F:
                    load = True
                    group_obj = F[self.group]
                    value = group_obj[item].attrs[attr]
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return value

    def attr_set(self, item, attr, value):
        '''\
        Set a value for an attribute

        Inputs:
            item: the dataset or group the attribute would belong to
            attr: the name of the attribute
            value: the value of the attribute
        '''

        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, 'r+') as F:
                    load = True
                    group_obj = F[self.group]
                    group_obj[item].attrs[attr] = value
            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)


    def attr_list(self, item):
        '''\
        List the attributes of a given item

        Inputs:
            item: the dataset or group the attribute would belong to
        '''
        n_try = 0
        load = False
        while (not load) and (n_try < self.retries):
            try:
                with h5py.File(self.fname, "r") as F:
                    load = True
                    group_obj = F[self.group]
                    attrs = list(group_obj[item].attrs)

            except OSError:
                n_try += 1
                #If the resource is busy, try again!
                time.sleep(np.random.random()*self.sleep)

        return attrs

    def attr_dict(self, item):
        '''\
        return a dictionary with all of the attributes for a given item

        Inputs:
            item: the dataset or group the attribute would belong to
        '''
        # Load all the keys
        keys = self.attr_list(item)
        # Initialize dictionary
        attrs = {}
        # Loop through each attr
        for key in keys:
            attrs[key] = self.attr_value(item, key)
        return attrs

    def attr_set_dict(self, item, attrs):
        '''\
        set each attribute from the dictionary

        Inputs:
            item: the dataset or group the attribute would belong to
            attrs: the dictionary of attributes
        '''
        # Loop through each attribute
        for key in attrs:
            # Set it
            self.attr_set(item, key, attrs[key])




'''Grids for direct evaluation fits
'''

class Grid(object):
    '''\
    Mesh object
    '''
    def __init__(
                 self,
                 Y,
                 lnL,
                 limits,
                 norm_limit=100.,
                ):
        '''Initialize a mesh object

        Parameters
        ----------
        Y: array like, shape = (npts,ndim)
            Input Samples to fit
        L: array like, shape = (npts,)
            Input sample likelihood values
        std: array like, shape = (ndim,)
            Input sample variance, useful for some things
        limits: array like, shape = (2,ndim)
            Input limits for space of mesh
        norm_limit: float (optional)
            Input maximum normalization constant, relative to guess
        '''
        # Imports
        from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
        import numpy as np
        # Exponentiate likelihood
        L = np.exp(lnL)
        # Downselect data
        keep = L > 0
        Y = Y[keep]
        L = L[keep]
        lnL = lnL[keep]
        # Normalize
        Lsum = np.sum(L)
        lnLsum = np.log(Lsum)
        # Save data
        self.Y = Y
        self.L = L#/Lsum
        self.L_norm = L/Lsum
        self.lnL = lnL# - lnLsum
        self.lnL_norm = lnL - lnLsum

        # Hold onto these
        self.npts = Y.shape[0]
        self.ndim = Y.shape[1]
        self.limits = limits
        # Core calculations
        self.mean = np.average(Y,weights=L,axis=0)
        self.cov = np.cov(Y.T,aweights=L)
        self.std = np.sqrt(np.diag(self.cov))
        # Derived calculations
        self._mu_scaled = self.mean/self.std
        self._cov_scaled = self.cov/np.outer(self.std,self.std)
        X_test = params_of_norm_mu_cov(1., self._mu_scaled, self._cov_scaled)
        self.norm_est = self.estimate_normalization_constant(X_test)
        self.norm_limit = norm_limit
        self.X_simple = params_of_norm_mu_cov(self.norm_est, self._mu_scaled, self._cov_scaled)

    def best_samples(self, n):
        ''' Find the best samples from our set

        Parameters
        ----------
        n: int
            Number of samples to find
        '''
        import numpy as np
        sort_indx = np.argsort(self.lnL)[::-1]
        best_indx = sort_indx[-n:]
        best_samp = self.Y[best_indx]
        best_L    = self.L[best_indx]
        best_lnL  = self.lnL[best_indx]
        return best_samp, best_L, best_lnL
        

    def estimate_normalization_constant(self, X_test, n=3):
        ''' Estimate the normalization constant for the fit

        Parameters
        ----------
        n: int
            Number of samples to use
        '''
        import numpy as np
        from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
        from gwalk.utils.multivariate_normal import multivariate_normal_pdf
        # Identify the highest likelihood point
        best_samp, best_L, best_lnL = self.best_samples(n)
        # Find the value of the multivariate normal distribution for the best points
        best_lnL_test = multivariate_normal_pdf(X_test, best_samp, scale=self.std, log_scale=True)
        norm_simple = np.mean(best_lnL - best_lnL_test)
        return norm_simple

    def polyfit_mu_sig_1d(self, x, y, sig_default, limits):
        '''\
        Return polynomial coefficients or best guess

        Parameters
        ----------
        x: array like, shape = (npts,)
            Input space values
        y: array like, shape = (npts,)
            Input function values
        sig_default: float
            Input default value of sigma
        limits: array like, shape = (2,)
            Input limits for x space
        '''
        # Imports 
        import numpy as np
        # Call polyfit
        a, b, c = np.polyfit(x,y,2)
        # check if a is viable
        if a < 0:
            # Use polynomial coefficients
            mu = -(0.5*b/a)
            sig = -0.5/a
        else:
            # Use maximum of data
            x = x[1:-1]
            y = y[1:-1]
            mu = x[np.argmax(y)]
            sig = sig_default
        # check limits
        if (mu < limits[0]) or (mu > limits[1]):
            # Use maximum of data
            x = x[1:-1]
            y = y[1:-1]
            mu = x[np.argmax(y)]

        return mu, sig

    ######## MultivariateNormal tools ########

    #### Call Constructor ####
    def construct_nal(
                      self,
                      seed=0,
                      sig_max=None,
                      labels=None,
                     ):
        ''' Construct a bounded multivariate normal model

        Parameters
        ----------
        seed: int, optional
            Input seed for random state
        sig_max: float, optional
            Input maximum sigma parameters, relative to scale
        '''
        # Imports
        import numpy as np
        from gwalk.model.parameter import Parameter
        from gwalk.bounded_multivariate_normal import MultivariateNormal
        # Initialize parameter list
        params = []
        # Loop parameters
        for i in range(self.ndim):
            # Pick a parameter guess
            guess = (self.limits[i][1] + self.limits[i][0])/2

            # Pick a label
            if labels is None:
                label = None
            else:
                label = labels[i]

            # Construct parameter 
            p = Parameter("p_%d"%i,guess,self.limits[i],label)
            params.append(p)

        pnorm = Parameter("norm", self.norm_est, [0,self.norm_limit*self.norm_est], "norm")

        # Construct Bounded Multivariate Normal object
        MV = MultivariateNormal(
                                params,
                                self.std,
                                seed,
                                sig_max,
                                norm=pnorm,
                               )
        self.X_lims = MV.read_limits()
        self.scale=MV.scale
        return MV

    def nal_save_kl(
                    self,
                    MV,
                    fname_nal,
                    label,
                    attrs=None,
                    mode='mean',
                    better='False',
                   ):
        '''Save MV object with kl divergence
        
        Parameters
        ----------
        MV: MultivaraiteNormal object
            Input bounded multivariate normal object
        fname_nal: str
            Input file location for nal fits
        label: str
            Input path to fit group
        attrs: dict, optional
            Input additional attributes to save with nal fit
        better: bool, optional
            Input save fit only if better?
        '''
        # Imports
        from gwalk.bounded_multivariate_normal import MultivariateNormal
        # Initialize attrs if none
        if attrs is None:
            attrs = {}
        # Get kl divergence
        attrs["kl"] = self.nal_kl_div(MV,MV.read_guess(),mode=mode).flatten()
        # Check for better
        if better:
            # Check if the fit already exists
            if MV.exists(fname_nal, label):
                # Load the existing fit
                MVexist = MultivariateNormal.load(fname_nal, label)
                # check MVexist[kl]
                kl_exist = self.nal_kl_div(MVexist,MVexist.read_guess(),mode=mode).flatten()
                # If kl_exist is lower than kl, return
                if kl_exist < attrs["kl"]:
                    return
        # Save fit
        MV.save(fname_nal, label, attrs=attrs)

    #### Convergence ####
    def nal_kl_div(
                   self,
                   MV,
                   X=None,
                  ):
        ''' Calculate the KL Divergence for a set of parameters

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts, nparams),optional
            Input test params for kl divergence
        '''
        # Imports 
        import numpy as np
        from relative_entropy_cython import relative_entropy_alt
        # Use built in value for X
        if X is None:
            X = MV.read_guess()
        ## Prep ##
        X = MV.check_sample(X)
        ## Calculate kl divergences ##
        # Get lnL_norm
        lnL_norm = MV.likelihood(
                                 self.Y,
                                 X=X,
                                 scale=MV.scale,
                                 log_scale=True,
                                )
        # Normalization of Q is handled by relative entropy alt
        #lnL_norm /= np.sum(lnL_norm)

        # Calculate the kl divergence
        kl = relative_entropy_alt(self.L_norm, self.lnL_norm, lnL_norm)
        return kl

    def nal_least_squares(
                          self,
                          MV,
                          X=None,
                          scalar=True,
                         ):
        ''' Calculate the least squares difference for a set of parameters

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts, nparams),optional
            Input test params for kl divergence
        scalar: bool (otpional)
            Input decide if you want a single number output instead of an array
        '''
        # Imports 
        import numpy as np
        # Use built in value for X
        if X is None:
            X = MV.read_guess()
        ## Prep ##
        X = MV.check_sample(X)
        keep = MV.satisfies_constraints(X)
        ## Calculate kl divergences ##
        # Get lnL_norm
        lnL = np.zeros((X.shape[0],self.npts))
        if np.sum(keep) > 0:
            lnL[keep] = MV.likelihood(
                                      self.Y,
                                      X=X[keep],
                                      scale=MV.scale,
                                      log_scale=True,
                                      lnL_offset=X[:,0],
                                     )
        lstsq = np.sum((self.lnL[None,:] - lnL)**2,axis=1)
        if scalar:
            lstsq = np.prod(lstsq)
        return lstsq


    def nal_kl_opt_fn(self,MV,k=-1.):
        ''' Return a function which evaluates the kl divergence given params

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        k: float, optional
            Input power of kl divergence
        '''
        import numpy as np
        def kl_div(X):
            kl = self.nal_kl_div(MV,X=X)
            return np.power(kl,k)
        return kl_div

    #### Guessing ####
    def nal_grid_guesses(
                         self,
                         MV,
                        ):
        ''' Fit the bounded multivariate normal model to the grid
        
        Parameters
        ----------
        MV: MultivariateNormal object
            Input some initialized Multivariate Normal object
        '''
        # Imports 
        import numpy as np
        from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
        # Identify number of guesses
        n_guess = 2
        # Initialize guesses
        Xg = np.tile(MV.read_guess(),(n_guess,1))
        Xg[1] = self.X_simple
        '''
        # Generate 1D evaluation guesses
        for i in range(self.ndim):
            y_test = y_test[keep].flatten()/MV.scale[i]
            L_test = np.log(L_test[keep].flatten())
            # Find maximum
            mu, sig = self.polyfit_mu_sig_1d(y_test, L_test, X[i+self.ndim], self.limits[i])
            Xg[0,i] = mu
            Xg[0,i+self.ndim] = sig

        # Generate 1D training guesses
        for i in range(self.ndim):
            # Load values
            y_train = self.marginals["1d_%d_x_train"%i]
            L_train = self.marginals["1d_%d_y_train"%i]
            bins = int(self.marginals["1d_%d_bins"%i])

            # Rescale training set 1
            y_train_1 = y_train[:bins].flatten()/MV.scale[i]
            L_train_1 = L_train[:bins].flatten()
            keep_1 = L_train_1 > 0
            y_train_1 = y_train_1[keep_1]
            L_train_1 = np.log(L_train_1[keep_1])
            # Rescale training set 2
            y_train_2 = y_train[bins:].flatten()/MV.scale[i]
            L_train_2 = L_train[bins:].flatten()
            keep_2 = L_train_2 > 0
            y_train_2 = y_train_2[keep_2]
            L_train_2 = np.log(L_train_2[keep_2])

            # fit training set 1
            mu, sig = self.polyfit_mu_sig_1d(y_train_1, L_train_1, X[i+self.ndim], self.limits[i])
            Xg[1,i] = mu
            Xg[1,i+self.ndim] = sig

            # fit training set 2
            mu, sig = self.polyfit_mu_sig_1d(y_train_2, L_train_2, X[i+self.ndim], self.limits[i])
            Xg[2,i] = mu
            Xg[2,i+self.ndim] = sig
        '''

        return Xg


    #### Fit methods ####

    def nal_fit_to_samples(self,MV,**kwargs):
        ''' Fit the bounded multivariate normal model to some samples
        Parameters
        ----------
        MV: MultivariateNormal object
            Input some initialized Multivariate Normal object
        '''
        # Imports
        import numpy as np
        from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
        # Generate simple guess
        Xs = params_of_norm_mu_cov(self.norm_est,self._mu_scaled,self._cov_scaled)
        # Fit to samples
        MV.assign_guess(Xs)
        return MV

    def nal_fit_scipy(self, MV, objective="kl", method="L-BFGS-B", **kwargs):
        ''' Fit the bounded multivariate normal model using scipy.optimize

        Parameters
        ----------
        MV: MultivariateNormal object
            Input some initialized Multivariate Normal object
        '''
        # Imports
        import numpy as np
        import scipy.optimize
        from gwalk.utils.multivariate_normal import mu_of_params
        from gwalk.utils.multivariate_normal import cov_of_params
        def kl_objective(X):
            kl = self.nal_kl_div(MV, X=X)
            return kl
        def lstsq_objective(X):
            lstsq = self.nal_least_squares(MV, X=X)
            return lstsq
            

        if objective == "kl":
            fn = kl_objective
        elif objective == "lstsq":
            fn = lstsq_objective
        else:
            raise RuntimeError("objective options: 'kl', 'lstsq'")

        MV.scipy_out = scipy.optimize.minimize(
                                               fn,
                                               self.X_simple,
                                               method=method,
                                               bounds=self.X_lims,
                                              )

        X_scipy = MV.scipy_out.x.flatten()
        if objective == "kl":
            # Still need a least squares fit for the normalization constant
            def lstsq_norm(X0):
                X_norm = np.copy(X_scipy)
                X_norm[0] = X0
                lstsq = self.nal_least_squares(MV, X=X_norm)
                return lstsq
            # Minimize this one
            norm_scipy_out = scipy.optimize.minimize(
                    lstsq_norm,X_scipy[0],method=method,bounds=[self.X_lims[0]])
            X_scipy[0] = norm_scipy_out.x.flatten()
                
        MV.assign_guess(X_scipy)
        return MV
            
            




# cython: infer_types=True
'''Calculate the Bhattacharyya distance of two distributions
'''
import cython
cimport cython
from cython.parallel import prange
from cython.cimports.libc.math import sqrt, log
import numpy as np 

######## C functions ########

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double _BC_coef(
                     double[::1] P,
                     double[::1] Q,
                     Py_ssize_t Psize,
                    ) nogil:
    '''Calculate the Bhattacharyya distance for two distributions

    Parameters
    ----------
    P: Memoryview C array, shape = (Psize)
        Input First distribution for comparison
    Q: Memoryview C array, shape = (Psize)
        Input Second distribution for comparison
    Psize: Py_ssize_t
        Input numper of points in domain
    '''
    cdef int k
    cdef double BC = 0
    for k in range(Psize):
        BC += sqrt(P[k] * Q[k])
    return BC

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double _BC_dist(
                     double[::1] P,
                     double[::1] Q,
                     Py_ssize_t Psize,
                    ) nogil:
    '''Calculate the Bhattacharyya distance for two distributions

    Parameters
    ----------
    P: Memoryview C array, shape = (Psize)
        Input First distribution for comparison
    Q: Memoryview C array, shape = (Psize)
        Input Second distribution for comparison
    Psize: Py_ssize_t
        Input numper of points in domain
    '''
    cdef int k
    cdef double BC = 0
    for k in range(Psize):
        BC += sqrt(P[k] * Q[k])
    BC = -log(BC)
    return BC

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double _H_dist(
                    double[::1] P,
                    double[::1] Q,
                    Py_ssize_t Psize,
                   ) nogil:
    '''Calculate the Hellinger distance for two distributions

    Parameters
    ----------
    P: Memoryview C array, shape = (Psize)
        Input First distribution for comparison
    Q: Memoryview C array, shape = (Psize)
        Input Second distribution for comparison
    Psize: Py_ssize_t
        Input numper of points in domain
    '''
    cdef int k
    cdef double H = 0 
    for k in range(Psize):
        H += sqrt(P[k] * Q[k])
    H = sqrt(1. - H)
    return H



######## General functions ########

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def bhattacharyya_coefficient(P,Q):
    '''Calculate the Bhattacharyya distance between two distributions'''
    # Handle inputs
    cdef Py_ssize_t Psize = P.size
    cdef Py_ssize_t Qnum
    cdef double max_value = np.finfo(np.double).max

    # Identify shapes
    if len(Q.shape) == 1:
        assert Q.size == Psize
        Q = Q.reshape((1,Psize))
        Qnum = 1
    else:
        assert Q.shape[1] == P.size
        Qnum = Q.shape[0]

    # Define vairables
    cdef int i, j
    Pc = np.asarray(P,order='c')
    Qc = np.asarray(Q,order='c')
    cdef double[::1] Pv = Pc
    cdef double[:,::1] Qv = Qc

    # Initialize output
    BC =  np.zeros(Qnum,dtype=np.double)
    cdef double[::1] BC_view = BC

    # Loop through Q
    for i in prange(Qnum,nogil=True):
        BC_view[i] = _BC_coef(Pv,Qv[i],Psize)

    return BC

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def bhattacharyya_distance(P,Q):
    '''Calculate the Bhattacharyya distance between two distributions'''
    # Handle inputs
    cdef Py_ssize_t Psize = P.size
    cdef Py_ssize_t Qnum
    cdef double max_value = np.finfo(np.double).max

    # Identify shapes
    if len(Q.shape) == 1:
        assert Q.size == Psize
        Q = Q.reshape((1,Psize))
        Qnum = 1
    else:
        assert Q.shape[1] == P.size
        Qnum = Q.shape[0]

    # Define vairables
    cdef int i, j
    Pc = np.asarray(P,order='c')
    Qc = np.asarray(Q,order='c')
    cdef double[::1] Pv = Pc
    cdef double[:,::1] Qv = Qc

    # Initialize output
    d =  np.zeros(Qnum,dtype=np.double)
    cdef double[::1] d_view = d

    # Loop through Q
    for i in prange(Qnum,nogil=True):
        d_view[i] = _BC_dist(Pv,Qv[i],Psize)

    return d

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def hellinger_distance(P,Q):
    '''Calculate the Hellinger distance between two distributions'''
    # Handle inputs
    cdef Py_ssize_t Psize = P.size
    cdef Py_ssize_t Qnum
    cdef double max_value = np.finfo(np.double).max

    # Identify shapes
    if len(Q.shape) == 1:
        assert Q.size == Psize
        Q = Q.reshape((1,Psize))
        Qnum = 1
    else:
        assert Q.shape[1] == P.size
        Qnum = Q.shape[0]

    # Define vairables
    cdef int i, j
    Pc = np.asarray(P,order='c')
    Qc = np.asarray(Q,order='c')
    cdef double[::1] Pv = Pc
    cdef double[:,::1] Qv = Qc

    # Initialize output
    d =  np.zeros(Qnum,dtype=np.double)
    cdef double[::1] d_view = d

    # Loop through Q
    for i in prange(Qnum,nogil=True):
        d_view[i] = _H_dist(Pv,Qv[i],Psize)

    return d



# cython: infer_types=True
'''Calculate the relative entropy using cython compiled functions

long description
'''
import cython
cimport cython
from cython.parallel import prange
from cython.cimports.libc.math import log, exp
import numpy as np

cdef extern from "math.h":
    double log(double x) nogil

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double _rel_entr(
                     double[::1] P,
                     double[::1] Q,
                     Py_ssize_t Psize,
                     double max_value,
                    ) nogil:
    '''Calculate the relative entropy between P and Q
    '''
    cdef int i
    cdef double r = 0
    for i in range(Psize):
        if (P[i] > 0) & (Q[i] > 0):
            r += P[i]*log(P[i]/Q[i])
        elif P[i] < 0:
            r = max_value
            return r
    return r

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double _rel_entr_alt(
                          double[::1] P,
                          double[::1] lnP,
                          double[::1] lnQ,
                          Py_ssize_t Psize,
                         ) nogil:
    '''Calculate the relative entropy between P and Q
    '''
    cdef int i
    cdef double r = 0
    cdef double Qc = 0
    cdef double lnQc
    # Calculate Q coefficient
    for i in range(Psize):
        Qc += exp(lnQ[i])
    lnQc = log(Qc)
    # Calculate relative entropy
    for i in range(Psize):
        r += P[i]*(lnP[i] - (lnQ[i] - lnQc))# - lnQc)
    return r

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def relative_entropy(P, Q,):
    '''Calculate the relative entropy between P and Q for many Q
    '''
    # Handle inputs
    cdef Py_ssize_t Psize = P.size
    cdef Py_ssize_t Qnum
    cdef double max_value = np.finfo(np.double).max

    # define variables
    cdef int i
    Pc = P.copy(order='c')
    Qc = Q.copy(order='c')
    cdef double[::1] Pv = Pc
    cdef double[:,::1] Qv = Qc

    # Identify sizes
    if len(Q.shape) == 1:
        assert Q.size == Psize
        Q = Q.reshape((1,Psize))
        Qnum = 1
    else:
        assert Q.shape[1] == Psize
        Qnum = Q.shape[0]

    # Initialize output
    r = np.zeros(Qnum,dtype=np.double)
    cdef double[::1] r_view = r

    # Loop through Q
    for i in prange(Qnum,nogil=True):
        r_view[i] = _rel_entr(Pv, Qv[i], Psize, max_value)

    return r

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def relative_entropy_alt(P, lnP, lnQ):
    '''Calculate the relative entropy between P and Q for many Q
    '''
    # Handle inputs
    cdef Py_ssize_t Psize = P.size
    cdef Py_ssize_t Qnum
    #Q = np.exp(lnQ)
    #Q /= np.sum(Q,axis=-1)[:,None]
    #lnQ = np.log(Q)


    # define variables
    cdef int i
    Pc = P.copy(order='c')
    lnPc = lnP.copy(order='c')
    lnQc = lnQ.copy(order='c')

    cdef double[::1] Pv = Pc
    cdef double[::1] lnPv = lnPc
    cdef double[:,::1] lnQv = lnQc

    # Identify sizes
    if len(lnQ.shape) == 1:
        assert lnQ.size == Psize
        lnQ = lnQ.reshape((1,Psize))
        Qnum = 1
    else:
        assert lnQ.shape[1] == Psize
        Qnum = lnQ.shape[0]

    # Initialize output
    r = np.zeros(Qnum,dtype=np.double)
    cdef double[::1] r_view = r

    # Loop through Q
    for i in prange(Qnum,nogil=True):
        r_view[i] = _rel_entr_alt(Pv, lnPv, lnQv[i], Psize)

    return r



#!/usr/bin/env python3
'''\
Provide infrastructure for models and hyperparameters
'''
#### Public Imports ####
import numpy as np
from maha_cython import mahalanobis_cython, pdf_exp_product
#### Globals ####
_INV_RT2 = 1./np.sqrt(2)
_LOG_2PI = np.log(2*np.pi)

#### General functions ####

def cov_of_std_cor(std,cor):
    '''Reconstruct the covariance from std and corelation
    Parameters
    ----------
    std: array like, shape = (npts, ndim)
        Input array of sigma values
    cor: array like, shape = (npts, ndim, ndim)
        Input array of correlation matrix values
    '''
    # Check dimensionality of std
    if len(std.shape) == 1:
        npts, ndim = 1, std.size
        std = std.reshape((npts,ndim))
    else:
        npts, ndim = std.shape
    # Check dimensionality of cor
    if len(cor.shape) == 2:
        cor = cor.reshape((npts,ndim,ndim))
    # Assert dimensionality match up
    assert cor.shape == ((npts,ndim,ndim))

    # DO NOT CHANGE THIS!
    std_expand = np.tensordot(std, np.ones(ndim), axes=0)
    var = std_expand * np.transpose(std_expand, axes=[0,2,1])
    # Reconstruct the covariance matrix
    cov = cor * var
    return cov

def std_of_cov(cov):
    '''Extract the vectorized std from the vectorized covariance
    Parameters
    ----------
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    std = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    return std

def cor_of_cov(cov):
    '''Reconstruct the vectorized correlation matrix from the covariance
    Parameters
    ----------
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    std = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    # Compute the inverse variance
    istd = 1./std
    # Compute the expansion of istd
    istd_expand = np.tensordot(istd, np.ones(ndim), axes=0)
    # Compute the inverse of the fully correlated covariance
    isig2 = istd_expand * np.transpose(istd_expand, axes=[0,2,1])
    # Calculate the correlation matrix
    cor = cov*isig2
    return cor


#### Functions that assume normal model object ####

## dimension checking/reduction ##
def n_param_of_ndim(ndim):
    '''Return the number of parameters associated witha guess in n dimensions
    Parameters
    ----------
    ndim: int
        Input number of dimensions
    '''
    return int((ndim*ndim + 3*ndim)//2) + 1

def ndim_of_n_param(n_param, max_dim=100):
    '''find the dimensionality of a guess
    Parameters
    ----------
    n_param: int
        Input number of parameters
    max_dim: int, optional
        Input maximum number of dimensions to try
    '''
    for i in range(max_dim):
        if n_param == n_param_of_ndim(i):
            return i
    raise RuntimeError("Failed to get number of dimensions for guess")

def params_reduce_1d(X, index):
    '''Reduce to a 1D marginal set of parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    index: int
        Input which dimension would we like to generate an evaluation set for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, n_param = X.shape
    ndim = ndim_of_n_param(n_param)
    # Protect against invalid choices
    assert ndim > index
    # Reduce dimensionality
    Xi = np.asarray([X[:,0], X[:,index + 1],X[:,index+ndim + 1]]).T
    return Xi

def params_reduce_2d(X, index, jndex):
    '''Reduce to a 2D marginal set of parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    index: int
        Input which dimension would we like to generate an evaluation set for
    jndex: int
        Input which other dimension would we like to generate an evaluation set for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, n_param = X.shape
    ndim = ndim_of_n_param(n_param)
    # Protect usage
    assert ndim > index
    assert index > jndex
    # Reduce dimensionality
    Xij = np.asarray([
                      X[:,0],
                      X[:,index + 1],
                      X[:,jndex + 1],
                      X[:,index+ndim + 1],
                      X[:,jndex+ndim + 1],
                      X[:,2*ndim + ((index*(index-1))//2) + jndex + 1]
                     ]).T
    return Xij

def params_reduce_dd(X, indices):
    '''Reduce to indexed marginals
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    indices: list
        Input list of indices we would like parameters for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, n_param = X.shape
    ndim = ndim_of_n_param(n_param)
    # Extract dimensionality of indices
    ndim_index = len(indices)
    n_param_index = n_param_of_ndim(ndim_index)
    # Initialize array
    Xp = np.empty((n_gauss, n_param_index))
    # Set normalization
    Xp[:,0] = X[:,0]
    # Loop through Xp dimensions
    for i in range(ndim_index):
         Xp[:,i+1] =              X[:,indices[i] + 1]
         Xp[:,i+ndim_index+1] =   X[:,indices[i]+ndim + 1]
    # Loop through correlation factors
    for i in range(ndim_index):
        for j in range(i):
            if indices[i] > indices[j]:
                index, jndex = indices[i], indices[j]
            elif indices[i] < indices[j]:
                index, jndex = indices[j], indices[i]
            else:
                raise RuntimeError("indices[i] == indices[j]")
            Xp[:,2*ndim_index + ((i*(i-1))//2) + j + 1] = \
                X[:,2*ndim + ((index*(index-1))//2) + jndex + 1]
    return Xp

                



## Conversions ##

def mu_of_params(X,scale=None,**kwargs):
    '''Reconstruct the vectorized mu parameters of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    if not (scale is None):
        if len(scale.shape) == 1:
            scale = scale[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    n_param = X.shape[1]
    ndim = ndim_of_n_param(n_param)
    # Reconstruct mu
    mu = X[:,1:ndim+1]
    if not (scale is None):
        mu*=scale
    return mu

def std_of_params(X,scale=None,**kwargs):
    '''Reconstruct the vectorized sigma parameters of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    if not (scale is None):
        if len(scale.shape) == 1:
            scale = scale[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    n_param = X.shape[1]
    ndim = ndim_of_n_param(n_param)
    # Reconstruct std
    std = X[:,ndim+1:2*ndim+1]
    if not (scale is None):
        std*=scale
    return std

def cor_of_params(X,**kwargs):
    '''Reconstruct the vectorized correlation matrix of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    n_param = X.shape[1]
    ndim = ndim_of_n_param(n_param)
    # Calculate std of params
    std = std_of_params(X)
    # Reconstruct correlation matrices
    cor = np.ones((n_gauss, ndim, ndim))*np.eye(ndim)
    # The correlation index is carefully kept consistent
    # DO NOT CHANGE THIS!
    cor_index = 2*ndim + 1
    for i in range(ndim):
        for j in range(i):
            cor[:,i,j] = cor[:,j,i] = X[:,cor_index]
            cor_index += 1
    if not cor_index == n_param:
        raise RuntimeError("Correlation Matrix reconstruction is broken")
    return cor

def cov_of_params(X,**kwargs):
    '''Reconstruct the vectorized corvariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    n_param = X.shape[1]
    ndim = ndim_of_n_param(n_param)
    # Calculate std of params
    std = std_of_params(X,**kwargs)
    # Calculate correlation matrix
    cor = cor_of_params(X,**kwargs)
    # Reconstruct the variance matrix
    # DO NOT CHANGE THIS!
    std_expand = np.tensordot(std, np.ones(ndim), axes=0)
    var = std_expand * np.transpose(std_expand, axes=[0,2,1])
    # Reconstruct the covariance matrix
    cov = cor * var
    return cov

def mu_cov_of_params(X,**kwargs):
    '''Reconstruct vectorized mu and covariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    mu = mu_of_params(X,**kwargs)
    cov = cov_of_params(X,**kwargs)
    return mu, cov

def params_of_mu_cov(mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    mu: array like, shape = (npts, ndim)
        Input Array of mu values
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if (len(mu.shape) == 1) and (len(cov.shape) == 2):
        mu = mu[None,:]
        cov = cov[None,:,:]
    elif (len(mu.shape) != 2) or (len(cov.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    # Identify useful information
    n_gauss = mu.shape[0]
    ndim = mu.shape[1]
    n_param = n_param_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (n_gauss, ndim, ndim):
        raise ValueError("cov is not shape (n_gauss, ndim, ndim)")
    # Initialize array
    X = np.empty((n_gauss,n_param))
    # Initialize normalization
    X[:,0] = 0.0
    # Insert mu values
    X[:,1:ndim+1] = mu
    # Insert std values
    X[:,ndim+1:2*ndim+1] = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    ## Compute Correlation ##
    cor = cor_of_cov(cov)
    ## Insert Correlation Values ##
    cor_index = 2*ndim +1
    for i in range(ndim):
        for j in range(i):
            X[:,cor_index] = cor[:,i,j]
            cor_index += 1
    if not cor_index == n_param:
        raise RuntimeError("Correlation matrix not extracted properly")
    # And we're done!
    return X

def params_of_norm_mu_cov(norm, mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    mu: array like, shape = (npts, ndim)
        Input Array of mu values
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if (len(mu.shape) == 1) and (len(cov.shape) == 2):
        mu = mu[None,:]
        cov = cov[None,:,:]
    elif (len(mu.shape) != 2) or (len(cov.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    # Identify useful information
    n_gauss = mu.shape[0]
    ndim = mu.shape[1]
    n_param = n_param_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (n_gauss, ndim, ndim):
        raise ValueError("cov is not shape (n_gauss, ndim, ndim)")
    # Initialize array
    X = np.empty((n_gauss,n_param))
    # Initialize normalization
    X[:,0] = norm
    # Insert mu values
    X[:,1:ndim+1] = mu
    # Insert std values
    X[:,ndim+1:2*ndim+1] = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    ## Compute Correlation ##
    cor = cor_of_cov(cov)
    ## Insert Correlation Values ##
    cor_index = 2*ndim +1
    for i in range(ndim):
        for j in range(i):
            X[:,cor_index] = cor[:,i,j]
            cor_index += 1
    if not cor_index == n_param:
        raise RuntimeError("Correlation matrix not extracted properly")
    # And we're done!
    return X

#### multivariate normal pdf ####
def multivariate_normal_pdf(
                            X,
                            Y,
                            scale = False,
                            log_scale = False,
                           ):
    '''Find the likelihood of some data with a particular guess of parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    Y: array like, shape = (npts, ndim)
        Input Array of samples to be evaluated
    scale: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
            if False: assume input data is PHYSICAL
            if True: assume input data is SCALED
            if (len(Y),) array: scale input by given values
            if (ngauss, len(Y)): Each sample gaussian has its own scale
    log_scale: bool, optional
        Input return log likelihood instead of likelihood?

    Outputs:
        L - array of likelihoods (n_gauss, n_pts)
    '''
    ## Imports ##
    # Public
    import numpy as np
    from scipy import stats

    ## Check Usage ##
    # Load data shapes
    X_shape = X.shape
    Y_shape = Y.shape

    # Intended use: X and Y are both rank 2
    if (len(X_shape) == 2) and (len(Y_shape) == 2):
        # Identify key shape values
        n_gauss = X_shape[0]
        n_param = X_shape[1]
        n_pts   = Y_shape[0]
        ndim   = Y_shape[1]
        # Check dimensionality
        if not n_param_of_ndim(ndim) == n_param:
            raise ValueError("Usage: X should be (n_guass, n_param)," + \
                                "Y should be (n_pts, ndim)")

    # If n_gauss is 1, we can work with that
    elif (len(X_shape) == 1) and (len(Y_shape) == 2):
        # Identify key shape values
        n_gauss = 1
        n_param = X_shape[0]
        n_pts   = Y_shape[0]
        ndim   = Y_shape[1]
        # Check dimensionality
        if not n_param_of_ndim(ndim) == n_param:
            raise ValueError("Usage: X should be (n_guass, n_param)," + \
                                "Y should be (n_pts, ndim)")
        else:
            # Correct dimensionality
            X = X[None,:]

    # If n_pts is 1, we can work with that
    elif (len(X_shape) == 2) and (len(Y_shape) == 1):
        # Identify key shape values
        n_gauss = X_shape[0]
        n_param = X_shape[1]
        n_pts   = 1
        ndim   = Y_shape[0]
        # Check dimensionality
        if not n_param_of_ndim(ndim) == n_param:
            raise ValueError("Usage: X should be (n_guass, n_param)," + \
                                "Y should be (n_pts, ndim)")
        else:
            # Correct dimensionality
            Y = Y[None,:]

    # If n_gauss is 1 and n_pts is 1, we can work with that
    elif (len(X_shape) == 1) and (len(Y_shape) == 1):
        # Identify key shape values
        n_gauss = 1
        n_param = X_shape[0]
        n_pts   = 1
        ndim   = Y_shape[0]
        # Check dimensionality
        if not n_param_of_ndim(ndim) == n_param:
            raise ValueError("Usage: X should be (n_guass, n_param)," + \
                                "Y should be (n_pts, ndim)")
        else:
            # Correct dimensionality
            X = X[None,:]
            Y = Y[None,:]

    # Other shapes are an error
    else:
        raise ValueError("Usage: X.shape = %s, Y.shape = %s"%(
                str(X_shape), str(Y_shape)))

    # Check scale
    if scale is False:
        # Scale not being used
        scale = np.ones((n_gauss, ndim))
    elif (len(scale.shape) == 1) and \
         (scale.shape[0] == ndim):
        # One scale fits all
        scale = np.ones((n_gauss, ndim)) * scale
    elif (len(scale.shape) == 2) and \
         (scale.shape[0] == n_gauss) and \
         (scale.shape[1] == ndim):
        # Dimensions are correct. Protect scope
        scale = scale.copy()
    else:
        raise ValueError("Usage: scale.shape = %s"%str(scale.shape))
        
    ## Reconstruct mu and covariance ##
    X_mu, X_cov = mu_cov_of_params(X,scale=scale)
    X_mu = np.asarray(X_mu,order='c')
    #XA = params_of_mu_cov(X_mu, X_cov)
    #print(np.max(np.abs(X - XA)))

    ## Eigenvector Decomposition ##
    # Find eigenvalues and eigenvectors (see scipy)
    s, u = np.linalg.eigh(X_cov)
    # Find the precision of the datatype used to store eigenvalues
    eps = stats._multivariate._eigvalsh_to_eps(s)
    # Initialize the log of the covariances
    d = np.zeros((n_gauss, ndim))
    # Find the eigenvalues which aren't zero
    keep = np.prod(s > eps, axis=1).astype(bool)

    # Find the log of the values which aren't zero
    d[keep] = np.log(s[keep])
    # Sum that quantity
    log_det_cov = np.sum(d, axis=1)

    # The goal of this section is to come up with a function U,
    # Such that ((x - mu) dot U)^2 is something like 
    # matmul((x - mu).T, matmul(Cov^-1, (x - mu)))
    # Note that this is done with arrays
    s_pinv = np.zeros_like(s)
    s_pinv[keep] = np.power(s[keep], -0.5)
    #raise Exception()
    U = u*s_pinv[:,None,:]

    # Calculate the mahalanobis factor
    maha = mahalanobis_cython(np.asarray(Y,order='c'), X_mu, U, n_gauss, n_pts, ndim)
    # Compute the likelihood!
    L = pdf_exp_product(maha, log_det_cov, ndim*_LOG_2PI, n_gauss, n_pts, ndim, log_scale=log_scale)
    L[~keep] = 0.

    return L

#### Helper Functions ####
def multivariate_normal_marginal1d(
                                   X,
                                   Y,
                                   index,
                                   scale = False,
                                  ):
    '''marginal1d evaluations
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    Y: array like, shape = (npts, ndim)
        Input Array of samples to be evaluated
    scale: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
            if False: assume input data is PHYSICAL
            if True: assume input data is SCALED
            if (len(Y),) array: scale input by given values
            if (ngauss, len(Y)): Each sample gaussian has its own scale
    index: int
        Input which dimension would we like to generate an evaluation set for
    '''
    # Assert dimensionality of data
    assert len(X.shape) == 2
    # Upack n_gauss and n_param
    n_gauss, n_param = X.shape
    # Retrieve ndim
    ndim = ndim_of_n_param(n_param)
    # Assert dimensionality of Y
    if len(Y.shape) == 2:
        if Y.shape[1] == ndim:
            Y = Y[:,index,None]
        elif Y.size == np.prod(Y.size):
            Y = Y.reshape(Y.size,1)
        else:
            raise RuntimeError("Y is some kind of weird shape")
    elif len(Y.shape) == 1:
        Y = Y.reshape(Y.size,1)
    else:
       raise RuntimeError("Y is some kind of weird shape")
        
    # Investigate scale
    if isinstance(scale,np.ndarray):
        if scale.size == ndim:
            scale = scale[index].flatten()
        elif scale.shape[1] == ndim:
            scale = scale[:,i].reshape((n_gauss,1))
    # Pick out X
    X = params_reduce_1d(X, index)
    L = multivariate_normal_pdf(X,Y,scale=scale)
    return L

def multivariate_normal_marginal2d(
                                   X,
                                   Y,
                                   index, jndex,
                                   scale = False,
                                  ):
    '''marginal1d evaluations
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    Y: array like, shape = (npts, ndim)
        Input Array of samples to be evaluated
    scale: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
            if False: assume input data is PHYSICAL
            if True: assume input data is SCALED
            if (len(Y),) array: scale input by given values
            if (ngauss, len(Y)): Each sample gaussian has its own scale
    index: int
        Input which dimension would we like to generate an evaluation set for
    jndex: int
        Input which other dimension would we like to generate an evaluation set for
    '''
    # Assert dimensionality of data
    assert len(X.shape) == 2
    # Upack n_gauss and n_param
    n_gauss, n_param = X.shape
    # Retrieve ndim
    ndim = ndim_of_n_param(n_param)
    # Assert dimensionality of Y
    assert len(Y.shape) == 2
    if Y.shape[1] == ndim:
        Y = Y[:,[index,jndex]]
    elif Y.shape[1] != 2:
        raise RuntimeError("Y is some kind of weird shape")
    # Investigate scale
    if isinstance(scale,np.ndarray):
        if scale.size == ndim:
            scale = scale[[index,jndex]].flatten()
        elif scale.shape[1] == ndim:
            scale = scale[:,[index,jndex]].reshape((n_gauss,2))
    # Pick out X
    X = params_reduce_2d(X, index, jndex)
    L = multivariate_normal_pdf(X,Y,scale=scale)
    return L



#!/usr/bin/env python3
'''Provide Relation support for a covariance matrix

Split into characteristic standard deviation and correlation factors

'''
######## Imports #########
from gwalk.model.parameter import Parameter
from gwalk.model.relation import Relation
from gwalk.utils.multivariate_normal import cov_of_std_cor
from gwalk.utils.multivariate_normal import std_of_cov
from gwalk.utils.multivariate_normal import cor_of_cov


class Covariance(Relation):
    '''It is useful to have a subclass of relationship for Covariance Matricies
    '''
    def __init__(
                 self,
                 ndim,
                 sig_max = 3.,
                 random_state=None,
                ):
        '''Initialize a relationship which represents a covaraince matrix

        This includes initializing parameters for each entree

        Inputs:
            ndim (int): The number of variables for the covariance matrix
            sig_max (float array): 
                Set a naive upper limit for the covariance matrix

        The covariance matrix will be made of (n^2 + n)/2 parameters,
            each representing a degree of freedom

        There are n degrees of freedom in the variance, sigma
        There are (n^2 - n)/2 degrees of freedom in the correlation matrix, rho

        Parameters
        ----------
        ndim: int
            Input The number of dimensions of data (not number of parameters)
        sig_max: float, optional
            Input The maximum allowed standard deviation for a relation
        random_state: np.RandomState
            Input numpy random state object for randomness
        '''
        ## Inputs ##
        # Public
        import numpy as np

        # Initialize a relationship
        super().__init__(random_state=random_state)
        # Define size
        self.ndim = ndim

        # Check sig_max
        if np.asarray(sig_max).size == 1:
            sig_max = np.ones((self.ndim))*sig_max
        elif np.asarray(sig_max).size == self.ndim:
            sig_max = np.asarray(sig_max).flatten()
        else:
            raise RuntimeError("sig_max must be scalar or 1d array of size n")

        # Generate Scale parameters
        # There will be n of these
        for i in range(self.ndim):
            # Name this entry
            tag = "std_%d"%(i)
            # Guess
            guess = 1.0
            # Guess limits
            limits = [1e-3, sig_max[i]]
            # Label
            label = r"$\sigma_{%d}$"%(i)

            # Initialize parameter
            p = Parameter(tag, guess, limits, label=label)
            self.add_parameter(p)

        # Generate Rho parameters
        for i in range(self.ndim):
            for j in range(i):
                # There are only half as many of these as CoV elements
                tag = "cor_%d_%d"%(i,j)
                # Guess these off diagonal elements to be uncorrelated
                guess = 0.
                # Limits are easy
                limits = [-1.,1.]
                # Label parameter
                label = r"$\rho_{%d,%d}$"%(i,j)

                # Initialize parameter
                p = Parameter(tag, guess, limits, label=label)
                self.add_parameter(p)


    def read_std(self):
        '''Return an std matrix with our current guess'''
        ## Imports ##
        # Public
        import numpy as np
        # Initialize std
        std = np.zeros((self.ndim,))
        # Loop through std parameters
        for i in range(self.ndim):
            # Generate tag
            tag = "std_%d"%(i)
            # Load value
            std[i] = self._parameters[tag].guess

        return std

    def read_cor(self):
        '''Return a correlation matrix with our current guesses'''
        ## Imports ##
        # Public
        import numpy as np
        # Initialize correlation
        cor = np.eye(self.ndim)
        # Loop through parameters
        for i in range(self.ndim):
            for j in range(i):
                # Generate tag
                tag = "cor_%d_%d"%(i,j)
                # Load value
                cor[i,j] = cor[j,i] = self._parameters[tag].guess

        return cor

    def read_cov(self):
        '''Return a covariance matrix with our current guesses'''
        ## Imports ##
        # Public
        import numpy as np
        # Load variance and correlation guesses
        std = np.atleast_2d(self.read_std())
        cor = self.read_cor()
        # Build covariance matrix
        var = std*std.T
        cov = cor.copy()*var
        return cov

    def std_of_params(self,X):
        '''Extract std parameters from samples

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample points for parameters
        '''
        ## Imports ##
        # Public
        import numpy as np
        # First check samples
        X = self.check_sample(X)

        ## Construct Variance ##
        # Initialize std
        std = np.zeros((X.shape[0],self.ndim,))
        # Loop through std parameters
        for i in range(self.ndim):
            # Generate tag
            tag = "std_%d"%(i)
            # Generate p_id
            p_id = self.p_map[tag]
            # Load value
            std[:,i] = X[:,p_id]

        return std

    def cor_of_params(self, X):
        '''Extract correlation matrix from samples

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample points for parameters
        '''
        ## Imports ##
        # Public
        import numpy as np
        # First check samples
        X = self.check_sample(X)

        # Initialize Correlation matrix
        cor = np.ones((X.shape[0], self.ndim, self.ndim))*np.eye(self.ndim)
        # Loop through correlation parameters
        for i in range(self.ndim):
            for j in range(i):
                # Generate tag
                tag = "cor_%d_%d"%(i,j)
                # Generate p_id
                p_id = self.p_map[tag]
                # Load value
                cor[:,i,j] = cor[:,j,i] = X[:,p_id]

        return cor

    def cov_of_params(self,X):
        '''Extract cov from samples

        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample points for parameters
        '''
        ## Imports ##
        # Public
        import numpy as np
        # First check samples
        X = self.check_sample(X)

        ## Construct Variance ##
        std = self.std_of_params(X)

        # Initialize Correlation matrix
        cor = self.cor_of_params(X)

        # Find the covariance
        cov = cov_of_std_cor(std, cor)
        return cov

    def std_of_cov(self, cov):
        '''Compute the std for an array of covariance matrixes

        Parameters
        ----------
        cov: array like, shape = (npts, ndim, ndim)
            Input covariance matrix
        '''
        return std_of_cov(cov)

    def cor_of_cov(self, cov):
        '''Compute the correlation for an array of covariance matrixes

        Parameters
        ----------
        cov: array like, shape = (npts, ndim, ndim)
            Input covariance matrix
        '''
        return cor_of_cov(cov)

    def params_of_cov(self, cov):
        '''Compute the list of params from a covariance matrix

        Parameters
        ----------
        cov: array like, shape = (npts, ndim, ndim)
            Input covariance matrix
        '''
        ## Imports ##
        # Public
        import numpy as np

        # Check dimensionality of cov
        if len(cov.shape) == 2:
            cov = np.asarray([cov])

        # Check the covariance matrix
        if not ((cov.shape[1] == self.ndim) and (cov.shape[2] == self.ndim)):
            raise RuntimeError("Invalid covariance matrix")
        
        # Initialize list of parameters
        X = np.empty((cov.shape[0], len(self._parameters)))
        
        # Compute variance and correlation
        std = std_of_cov(cov)
        cor = cor_of_cov(cov)
    
        # Loop through the std parameters
        for i in range(self.ndim):
            # Identify the std tag
            tag = "std_%d"%(i)
            # Identify parameter id
            p_id = self.p_map[tag]
            # Assign std parameters
            X[:,p_id] = std[:,i]

        # Loop through correlation parameters
        for i in range(self.ndim):
            for j in range(i):
                # Identify the correlation tag
                tag = "cor_%d_%d"%(i,j)
                # Identify parameter id
                p_id = self.p_map[tag]
                # Assign correlation parameters
                X[:,p_id] = cor[:,i,j]

        # Check samples
        X = self.check_sample(X)

        return X

    def satisfies_constraints(self, X):
        ''' Check if covariance Relation constraints are satisfied

        A Covariance Matrix must have positive Eigenvalues
        
        Parameters
        ----------
        X: array like, shape = (npts, nd)
            Input sample points for parameters
        '''
        ## Inputs ##
        # Public
        import numpy as np
        from numpy.linalg import eigvals, eigvalsh
        # First check samples
        X = self.check_sample(X)

        ## Construct Covariance Matrix ##
        cov = self.cov_of_params(X)

        # Find lower bound
        t = cov.dtype.char.lower()
        factor = {'f':1E3, 'd':1E6}
        cond = factor[t] * np.finfo(t).eps
        eps = cond * np.max(abs(cov))

        # Make sure variance guesses make sense
        c1 = super().satisfies_constraints(X)
        # Find if all the eigenvalues are positive
        c2 = np.bitwise_and.reduce(np.atleast_2d(eigvalsh(cov) > eps),axis=1)
        # Find the valid samples
        constrained_indx = c1 & c2
        return constrained_indx
                
